function [U,A,V] = new_TSB(A)
    % new_TSB: Function to transform a matrix A into a bidiagonal matrix
    % Inputs:
    %   A - Input matrix to be transformed into a bidiagonal matrix.
    % Outputs:
    %   U, A, V - Matrices such that A is the transformed bidiagonal matrix.

    % Check if any input is given
    if nargin==0
        % If no input, generate a random 100x60 matrix
        A = rand(100,60);
        D = A;
    end

    % Get dimensions of A
    [m,n] = size(A);                              

    % Transpose A if it has more columns than rows
    if m < n                                      
        A = A';
    end

    % Update size after possible transpose
    [m,n] = size(A);

    % Compute the optimal k value for transformation
    k = (2*n - m) * (2*n > m);                     

    % Perform GKB transformation
    [H,A,H_t] = GKB(A,k);                         

    % Perform LHC transformation
    [Q1,A,Q2] = LHC(A,k);                          

    % Compute left multiplication matrix
    U = H * Q1;                                    

    % Compute right multiplication matrix
    V = Q2 * H_t;                                  
end

function [H,A,H_t] = GKB(A,k)
    % GKB: Function to transform the first k rows and columns of A into a bidiagonal matrix
    % Outputs:
    %   A - Matrix with first k rows and columns transformed into a bidiagonal matrix.
    %   H - Left multiplication matrix.
    %   H_t - Right multiplication matrix.

    % Check if any input is given
    if nargin==0
        % If no input, generate a random 10x7 matrix and set k=5
        A = rand(10,7);
        k = 5;
    end

    % Initialize identity matrices for transformation
    [m,n] = size(A);
    H = eye(m);
    H_t = eye(n);

    % Return early if k is 0
    if k == 0
        return
    end

    % Initialize matrices for storing v and beta
    V_t = zeros(n,n);
    V = zeros(m+1,n);

    % Main loop for GKB transformation
    for i = 1:k                                                      
        % Define temporary matrices for transformation
        [v,beta,sigma] = house2(A(i:m,i));                           

        % Apply fast Householder transformation with error reduction
        A(i,i) = sigma;                                               
        A(i+1:m,i) = 0;
        A(i:m,i+1:n) = A(i:m,i+1:n) - v * (v' * A(i:m,i+1:n)) * beta;

        % Store v and beta
        V(i:m,i) = v;
        V(m+1,i) = beta;

        % Check if row transformation is needed
        if i+2 <= n                                                  
            [v,beta,sigma] = house2(A(i,i+1:n)');                     

            % Apply transformation to the rows
            A(i,i+1) = sigma;
            A(i,i+2:n) = 0;
            A(i+1:m,i+1:n) = A(i+1:m,i+1:n) - (A(i+1:m,i+1:n) * v) * v' * beta;
            V_t(i:n-1,i) = v;
            V_t(n,i) = beta;
        end
    end

    % Apply transformations to H and H_t matrices
    for i = 1:k
        H(:,i:m) = H(:,i:m) - H(:,i:m) * V(i:m,i) * V(i:m,i)' * V(m+1,i);
        if i+2 <= n
            H_t(i+1:n,:) = H_t(i+1:n,:) - V_t(i:n-1,i) * (V_t(i:n-1,i)' * H_t(i+1:n,:)) * V_t(n,i);
        end
    end
end

function [Q1,A,H_t] = LHC(A,k)
    % LHC: Function to transform rows and columns of A beyond k+1 into a bidiagonal matrix
    % Outputs:
    %   A - Matrix with rows and columns beyond k+1 transformed into a bidiagonal matrix.
    %   Q1 - Left multiplication matrix.
    %   H_t - Right multiplication matrix.

    % Check if any input is given
    if nargin==0
        % If no input, generate a random 10x7 matrix and set k=0
        A = rand(10,7);
        k = 0;
    end

    % Get the size of A
    [m,n] = size(A);
    Q1 = eye(m);
    H = eye(m);
    H_t = eye(n);

    % Return early if no transformation is needed
    if k == n
        return
    end

    % Update A and Q1 matrices with right QR transformation
    [Q1(k+1:m,k+1:m), A(k+1:m,k+1:n)] = qr_right(A(k+1:m,k+1:n));

    % Initialize matrices for transformations
    V = zeros(n,n);
    V_t = zeros(n,n);

    % Check if row transformation is needed
    if k+2 < n
        % Main loop for LHC transformation
        for i = k+1:n-2
            % Apply transformation to rows
            [v,beta,sigma] = house2(A(i,i+1:n)');
            A(i,i+1) = sigma;
            A(i,i+2:n) = 0;
            A(i+1:n,i+1:n) = A(i+1:n,i+1:n) - (A(i+1:n,i+1:n) * v) * v' * beta;
            V_t(i:n-1,i) = v;
            V_t(n,i) = beta;

            % Apply transformation to columns
            [v,beta,sigma] = house2(A(i+1:n,i+1));
            A(i+1,i+1) = sigma;
            A(i+2:n,i+1) = 0;
            A(i+1:n,i+2:n) = A(i+1:n,i+2:n) - v * (v' * A(i+1:n,i+2:n)) * beta;
            V(i:n-1,i) = v;
            V(n,i) = beta;
        end

        % Apply final transformations
        for i = k+1:n-2
            H_t(i+1:n,:) = H_t(i+1:n,:) - V_t(i:n-1,i) * (V_t(i:n-1,i)' * H_t(i+1:n,:)) * V_t(n,i);
            H(:,i+1:n) = H(:,i+1:n) - H(:,i+1:n) * V(i:n-1,i) * V(i:n-1,i)' * V(n,i);
        end
    end

    % Combine Q1 and H matrices
    Q1 = Q1 * H;
end

function [Q, R] = qr_right(A)
    % qr_right: Function to perform QR factorization using Householder transformations
    % Inputs:
    %   A - Input matrix to be factorized.
    % Outputs:
    %   Q - Orthogonal matrix from the QR factorization.
    %   R - Upper triangular matrix from the QR factorization.

    % Get the size of A
    [m,n] = size(A);

    % Initialize matrices for storing Householder vectors and Beta values
    V = zeros(m,n);
    Beta = zeros(1,n);

    % Loop over each column of A for Householder transformations
    for col = 1:n
        % Compute Householder vector and Beta for the current column
        [v, beta] = house2(A(col:m, col)); 

        % Store the Householder vector and Beta
        V(col:m, col) = v;
        Beta(col) = beta;

        % Apply the Householder transformation to the current column
        A(col:m, col:n) = A(col:m, col:n) - v * (beta * (v' * A(col:m, col:n)));
    end

    % Extract the upper triangular matrix R
    R = triu(A, 0);

    % Initialize the orthogonal matrix Q
    Q = eye(m) - Beta(1) * V(:, 1) * V(:, 1)';

    % Construct Q using the Householder vectors
    for col = 2:n
        Q(:, col:end) = Q(:, col:end) - Q(:, col:end) * V(col:end, col) * (Beta(col) * V(col:end, col)');
    end
end
