function [u,d,v] = ouyang_SVD(A)
    % ouyang_SVD: Custom function to compute the Singular Value Decomposition (SVD)
    % Inputs:
    %   A - Input matrix to decompose.
    % Outputs:
    %   u, d, v - Matrices representing the SVD of A (A = u*d*v').

    % Get the size of A
    [m,n] = size(A);

    % Compute eigenvalues and eigenvectors of A*A' using deflation and QR shift
    [d,u] = deflation_qr_shift(A);

    % Convert eigenvalues to singular values
    d = sqrt(d);

    % Compute the matrix V
    for i = 1:n
        v(:,i) = A' * u(:,i) / d(i,i); % Calculate each column of V
    end
end

function [lambda,hatQ] = deflation_qr_shift(A)
    % deflation_qr_shift: Function to compute eigenvalues and eigenvectors using deflation and QR shift
    % Inputs:
    %   A - Input matrix for which eigenvalues and eigenvectors are calculated.
    % Outputs:
    %   lambda - Diagonal matrix of eigenvalues.
    %   hatQ - Matrix of eigenvectors.

    % Check if any input is given
    if nargin == 0
        % Generate a test case if no input
        a = rand(10,1);
        b = rand(9,1);
        A = diag(a) + diag(b,1);
        A = A' * A; % Make A symmetric
    end

    % Multiply A by its transpose
    A = A * A';

    % Get the number of columns of A
    [~,n] = size(A);

    % Initialize matrices for eigenvalues and eigenvectors
    lambda = zeros(n,1);
    hatQ = eye(n,n);

    % Main loop for QR shift algorithm
    for m = n:-1:2
        % Continue until the off-diagonal element is smaller than epsilon
        while abs(A(m,m-1)) > eps
            % Apply Wilkinson's shift
            B = A(m-1:m,m-1:m);
            
            % Compute the eigenvalue closest to A(m,m)
            e = myeig(B,A(m,m));

            % Perform QR decomposition with Givens rotation
            [Q,R] = qrgivens(A-e*eye(m));
            A = R*Q + e*eye(m);

            % Keep A tridiagonal
            A = diag(diag(A,0),0) + diag(diag(A,1),1) + diag(diag(A,-1),-1);

            % Update eigenvectors
            hatQ(:,1:m) = hatQ(:,1:m) * Q;
        end

        % Store the eigenvalue
        lambda(m) = A(m,m);

        % Remove the last column and row of A
        A(:,m) = [];
        A(m,:) = [];
    end

    % Store the last eigenvalue
    lambda(1) = A(1,1);

    % Combine eigenvalues and eigenvectors, and sort in descending order
    L = [lambda, hatQ'];
    L = sortrows(L, 'descend');

    % Extract sorted eigenvalues and eigenvectors
    lambda = diag(L(:,1));
    hatQ = L(:,2:n+1)';
end

function [e] = myeig(A,k)
    % myeig: Function to compute the eigenvalue of a 2x2 matrix closest to a given value k
    % Inputs:
    %   A - A 2x2 matrix.
    %   k - The reference value to which the closest eigenvalue is computed.
    % Outputs:
    %   e - The eigenvalue of matrix A that is closest to k.

    % Compute the mean of the diagonal elements of A
    m = (A(1,1) + A(2,2)) / 2;

    % Compute the product of diagonal elements minus the product of off-diagonal elements
    p = A(1,1)*A(2,2) - A(1,2)*A(2,1);

    % Calculate the distance from the mean to the real eigenvalues
    dist = sqrt(m*m - p);

    % Determine the closest eigenvalue to k
    if m - k > 0
        e = m - dist;  % Choose the smaller eigenvalue if the mean is larger than k
    else
        e = m + dist;  % Choose the larger eigenvalue if the mean is smaller than k
    end
end
