function [Q,A] = qrgivens(A)
    % qrgivens: Function to perform QR decomposition using Givens rotation
    % Inputs:
    %   A - Input square matrix for QR decomposition.
    % Outputs:
    %   Q - Orthogonal matrix from QR decomposition.
    %   A - Upper triangular matrix (R) from QR decomposition.

    % Get the size of A
    [m,~] = size(A);

    % Initialize the orthogonal matrix Q as an identity matrix
    Q = eye(m);

    % Initialize a 2x2 matrix for Givens rotation
    G = zeros(2,2);

    % Main loop for QR decomposition
    for j = 1:m-2
        i = j + 1;

        % Compute the Givens rotation for elements A(j,j) and A(i,j)
        [c,s] = givensrotation(A(j,j), A(i,j));

        % Construct the Givens rotation matrix
        G = [c -s; s c];

        % Apply the Givens rotation to a part of matrix A
        A(j:i,j:j+2) = G' * A(j:i,j:j+2);

        % Eliminate the off-diagonal element
        A(i,j) = 0;

        % Update the orthogonal matrix Q
        Q(:,j:i) = Q(:,j:i) * G;
    end

    % Handle the last column separately to prevent out-of-bounds errors
    [c,s] = givensrotation(A(m-1,m-1), A(m,m-1));
    G = [c -s; s c];
    A(m-1:m,m-1:m) = G' * A(m-1:m,m-1:m);
    A(m,m-1) = 0;
    Q(:,m-1:m) = Q(:,m-1:m) * G;
end

function [c,s] = givensrotation(a,b)
    % givensrotation: Compute the Givens rotation matrix components
    % Inputs:
    %   a, b - Elements of the matrix to be rotated.
    % Outputs:
    %   c, s - Cosine and sine components of the Givens rotation matrix.

    % Check if the off-diagonal element is zero
    if b == 0
        c = 1;
        s = 0;
    else
        % Choose computation based on the magnitude of a and b
        if abs(b) > abs(a)
            r = a / b;
            s = 1 / sqrt(1 + r^2);
            c = s * r;
        else
            r = b / a;
            c = 1 / sqrt(1 + r^2);
            s = c * r;
        end
    end
end
